import matplotlib.pyplot as plt
import numpy as np


def plot_different_cell_type(
        dict_time_series, cell_types_list, time_step=1, figure_title='',
        cell_types_names=[]):
    """
    plot 1d chart of the count of cell types at different time points

    Parameters
    ----------
    dict_time_series: dict
        a dictionary that contains number of occurrence for each cell type
        throughout different time points
    cell_types_list: list
        a list of different type of cell that we want to count. E.g. [1, 2, 3].
    time_step: int (default=1)
        the time step between time points
    figure_title: str, optional (default=None)
        the title for the plotted figure
    cell_types_names: list, optional (default is cell_types_list)
        the name of different cell type. If not specified, the names will be
        taken from the cell_types_list

    Returns
    -------
    ax: matplotlib axis
        axis of the plot
    """
    if not dict_time_series and not isinstance(dict_time_series, dict):
        print("the provided dictionary is not in the required format")
        return -1
    if not isinstance(cell_types_list, list):
        print("the provided cell types list is not in the required format")
        return -1
    if not isinstance(time_step, int):
        print("the time step should be in int")
        return -1
    if not cell_types_names:
        cell_types_names = cell_types_list

    _, ax = plt.subplots()
    # get all time points
    time_interval = list(dict_time_series.keys())
    y_list = []
    # create empty neasted lists for each cell type
    for i in range(len(cell_types_list)):
        y_list.append(list())
    # iterate through the different time interval and save the cells data in
    # the neasted lists
    for key, data_dataframe in dict_time_series.items():
        for i in range(len(cell_types_list)):
            y_list[i].append(data_dataframe.iloc[i]['n_cells'])
    # plot the change of cell types throughout different time points
    ax.set_ylabel('Number of cells')
    ax.set_xlabel('Time (h)')
    ax.set_title(figure_title)
    for i in range(len(cell_types_list)):
        ax.plot(time_interval, y_list[i])
        # set up time points
    ax.set_xticks(np.arange(time_interval[0], time_interval[-1], time_step))
    ax.legend(cell_types_names)
    return ax


def plot_cluster_size_time_point(dict_time_series, figure_title=''):
    """
    plot a bar chart of the size of clusters at single time point

    Parameters
    ----------
    dict_time_series: dict
        a dictionary that contains number of occurrence for each cell type
        throughout different time points
    figure_title: str, optional (default=None)
        the title for the plotted figure

    Returns
    -------
    ax: matplotlib axis
        axis of the plot
    """
    if not dict_time_series and not isinstance(dict_time_series, dict):
        print("the provided dictionary is not in the required format")
        return -1
    _, ax = plt.subplots()
    # get all time points
    y_list = []
    # create empty neasted lists for each cell type
    keys_list = []
    # iterate through clusters in the dict_time_series
    for key, data_dict in dict_time_series.items():
        y_list.append(len(data_dict))
        keys_list.append(key)
    ax.bar(keys_list, y_list)
    # set up the label of x and y axis
    ax.set_ylabel('Size of cluster')
    ax.set_xlabel('Cluster index number')
    # set the title of the bar chart
    ax.set_title(figure_title)
    keys_list = list(map(str, keys_list))
    ax.set_xticks(keys_list)
    return ax


def plot_cluster_count_all_time_point(
        dict_time_series, time_step=1, figure_title=''):
    """
    plot a bar chart of the counts of clusters at multiple time points

    Parameters
    ----------
    dict_time_series: dict
        a dictionary that contains number of occurrence for each cell type
        throughout different time points
    time_step: int (default=1)
        the time step between time points
    figure_title: str, optional (default=None)
        the title for the plotted figure

    Returns
    -------
    ax: matplotlib axis
        axis of the plot
    """
    if not dict_time_series and not isinstance(dict_time_series, dict):
        print("the provided dictionary is not in the required format")
        return -1
    _, ax = plt.subplots()
    # get the cluster counts
    y_list = list(dict_time_series.values())
    # get different time points
    keys_list = list(dict_time_series.keys())
    ax.bar(keys_list, y_list)
    # set up the label of x and y axis
    ax.set_ylabel('Number of clusters')
    ax.set_xlabel('time (h)')
    # set the title of the bar chart
    ax.set_title(figure_title)
    ax.set_xticks(np.arange(keys_list[0], keys_list[-1], time_step))
    return ax


def plot_cluster_size_all_time_point(
        dict_time_series, time_step=1, figure_title=''):
    """
    plot a bar chart of the size of clusters at multiple time points

    Parameters
    ----------
    dict_time_series: dict
        a dictionary that contains number of occurrence for cluster index
        throughout different time points
    time_step: int (default=1)
        the time step between time points
    figure_title: str, optional (default=None)
        the title for the plotted figure

    Returns
    -------
    ax: matplotlib axis
        axis of the plot
    """
    if not dict_time_series and not isinstance(dict_time_series, dict):
        print("the provided dictionary is not in the required format")
        return -1
    _, ax = plt.subplots()
    # get the cluster counts
    y_list = list(dict_time_series.values())
    # get different time points
    keys_list = list(dict_time_series.keys())
    keys_list_str = [str(x) for x in keys_list]
    ax.bar(keys_list_str, y_list)
    # set up the label of x and y axis
    ax.set_ylabel('Number of clusters')
    ax.set_xlabel('Cluster index')
    # set the title of the bar chart
    ax.set_title(figure_title)
    ax.set_xticks(keys_list_str)
    return ax


def plot_active_cell_all_time_point(
        dict_time_series, time_step=1, figure_title=''):
    """
    plot a bar chart of the number of active infected cell at multiple time
    points

    Parameters
    ----------
    dict_time_series: dict
        a dictionary that contains the total number of active infected cells at
        different time points
    time_step: int (default=1)
        the time step between time points
    figure_title: str, optional (default=None)
        the title for the plotted figure

    Returns
    -------
    ax: matplotlib axis
        axis of the plot
    """
    if not dict_time_series and not isinstance(dict_time_series, dict):
        print("the provided dictionary is not in the required format")
        return -1
    _, ax = plt.subplots()
    # get the cluster counts
    y_list = list(dict_time_series.values())
    # get different time points
    keys_list = list(dict_time_series.keys())
    ax.bar(keys_list, y_list)
    # set up the label of x and y axis
    ax.set_ylabel('Number of CC contributor cells')
    ax.set_xlabel('time (h)')
    # set the title of the bar chart
    ax.set_title(figure_title)
    ax.set_xticks(keys_list)

    return ax



import os
import logging
import pandas as pd
from typing import Callable, Tuple, Union
import pyabc
import numpy as np
from ..model import MorpheusModel, MorpheusModels
from numbers import Number

logger = logging.getLogger("FitMultiCell.PEtab_MS")

try:
    import petab_MS
    import petab_MS.C as C
except ImportError:
    petab = C = None
    logger.error("Install petab_MS (see https://github.com/icb-dcm/petab) "
                 "to use the petab_MS functionality.")


# priors that are evaluated "on-scale"
SCALED_PRIORS = [C.PARAMETER_SCALE_UNIFORM,
                 C.PARAMETER_SCALE_NORMAL,
                 C.PARAMETER_SCALE_LAPLACE]


class PetabImporter():
    """Import a PEtab model to parameterize it using FitMultiCell pipeline.
    This class provides methods to generate prior, Morpheus, par_maps
    for a pyABC analysis.
<<<<<<< HEAD

=======
>>>>>>> 0acb391d85cb289fd75d9320e1aaf4bf99d1243f
    Parameters
    ----------
    petab_problem:
        A PEtab problem containing all information on the parameter estimation
        problem.
    """
    def __init__(
            self,
            petab_problem: petab_MS.Problem):
        self.petab_problem = petab_problem

    def create_model(self):
        """Create a model.

        Returns
        -------
        Morpheus_model: A model of a class MorpheusModels
        """
        return create_model(self)

    def create_prior(self) -> pyabc.Distribution:
        """Create prior.

        Note: The returned parameters are on the `objectivePriorType` scale,
        i.e. on `parameterScale` if the prior is one of `parameterScale...`,
        otherwise on linear scale.
        The model must then take care of transforming the parameters from
        prior scale to the right scale.
        Effectively, pyABC thus ignores the `parameterScale`. If this is an
        issue as sampling e.g. on lin or log scale could make a difference,
        this can be revised.

        Returns
        -------
        prior: A valid pyabc.Distribution for the parameters to estimate.
        """
        return create_prior(parameter_df=self.petab_problem.parameter_df)

    def get_par_map(self) -> dict:
        """Create parameter map to be use with the Morpheus model.
        Returns
        -------
        par_map: A dictionary with the xpath for all parameters.
        """
        return get_par_map(parameter_df=self.petab_problem.parameter_df)

    def get_nominal_parameters(
            self, target_scale: str = 'prior') -> pyabc.Parameter:
        """Get nominal parameters.
        Parameters
        ----------
        target_scale:
            Scale on which to return the parameters.
            Values: 'prior'|'scaled'|'lin'.
            If 'prior', they are on the scale the corresponding prior is on.
            If 'scaled', they are scaled, if 'lin', they are unscaled.
        Returns
        -------
        par_nominal:
            The nominal parameters if provided in the petab parameters data
            frame.
        """
        return get_nominal_parameters(
            parameter_df=self.petab_problem.parameter_df,
            target_scale=target_scale,
            prior_scales=self.prior_scales,
            scaled_scales=self.scaled_scales,
        )

    def get_bounds(
            self, target_scale: str = 'prior', use_prior: bool = True) -> dict:
        """Get bounds.
        Parameters
        ----------
        target_scale:
            Scale to get the nominal parameters on.
            Can be 'lin', 'scaled', or 'prior'.
        use_prior: Whether to use prior overrides if of type uniform.
        Returns
        -------
        bounds: Dictionary with a (lower, upper) tuple per parameter.
        """
        return get_bounds(
            parameter_df=self.petab_problem.parameter_df,
            target_scale=target_scale,
            prior_scales=self.prior_scales,
            scaled_scales=self.scaled_scales,
            use_prior=use_prior,
        )

    def get_parameter_names(self, target_scale: str = 'prior'):
        """Get meaningful parameter names, corrected for target scale."""
        parameter_df = petab_MS.normalize_parameter_df(
            self.petab_problem.parameter_df)

        # scale
        if target_scale == C.LIN:
            target_scales = {key: C.LIN for key in self.prior_scales}
        elif target_scale == 'prior':
            target_scales = self.prior_scales
        elif target_scale == 'scaled':
            target_scales = self.scaled_scales
        else:
            raise ValueError(f"Did not recognize target scale {target_scale}")

        names = {}
        for _, row in parameter_df.reset_index().iterrows():
            if row[C.ESTIMATE] == 0:
                continue
            key = row[C.PARAMETER_ID]
            name = str(key)
            if C.PARAMETER_NAME in parameter_df:
                if not petab_MS.is_empty(row[C.PARAMETER_NAME]):
                    name = str(row[C.PARAMETER_NAME])

            target_scale = target_scales[key]
            if target_scale != C.LIN:
                # mini check whether the name might indicate the scale already
                if not name.startswith("log"):
                    name = target_scale + "(" + name + ")"
            names[key] = name
        return names

    def get_exp_cond_map(self) -> dict:
        """get a dictionary of the experimental conditions.
        Returns
        -------
        condition_dict: A dictionary of for the xpath and the value for each
        experimental conditions.
        """

        return get_exp_cond_map(condition_df=self.petab_problem.condition_df)

    def import_petab_problem(self):
        """
        Import model from PEtab problem.
        :return:
            The imported model.
        """
        return import_petab_problem(self.petab_problem)

    def _sanity_check(self):
        """Some checks to identify potential problems."""
        if any(self.scaled_scales[key] != C.LIN
               and self.prior_scales[key] == C.LIN
               for key in self.prior_scales):
            logger.warning(
                "Found parameters with prior scale lin, parameter scale not "
                "lin. Note that pyABC currently ignores the parameter scale "
                "in this case and just performs sampling on the prior scale.")


def create_model(importer):
    conditions = importer.petab_problem.condition_df
    cond_status = get_nr_exp_condition(conditions)
    par_map_imported = importer.get_par_map()
    par_scale = importer.petab_problem.get_optimization_parameter_scales()

    if cond_status == "no_condition":
        model_con1 = MorpheusModel(
            importer.petab_problem.model_file, par_map=par_map_imported,
            par_scale=par_scale, timeout=100)
        return model_con1
    elif cond_status == "one_condition":
        exp_cond_imported = importer.get_exp_cond_map()
        model_con1 = MorpheusModel(
            importer.petab_problem.model_file, par_map=par_map_imported,
            exp_cond_map={'condition1': exp_cond_imported['condition1']},
            par_scale=par_scale, timeout=100)
        return model_con1
    else:
        exp_cond_imported = importer.get_exp_cond_map()
        models_list = []
        for cond_key, cond_val in exp_cond_imported.items():
            tmp_model = MorpheusModel(
                importer.petab_problem.model_file, par_map=par_map_imported,
                exp_cond_map={cond_key: cond_val},
                par_scale=par_scale, timeout=100)
            models_list.append(tmp_model)

        return MorpheusModels(models_list=models_list)


# =======
#            petab_problem: petab_MS.Problem):
#        self.petab_problem = petab_problem
#        self.prior_scales, self.scaled_scales = get_scales(
#            self.petab_problem.parameter_df)
#        self._sanity_check()
# >>>>>>> ece732b (add PEtab support ot FMC)


def create_prior(parameter_df: pd.DataFrame) -> pyabc.Distribution:
    """Create prior.
    Note: The prior generates samples according to
    `OBJECTIVE_PRIOR_TYPE` and `OBJECTIVE_PRIOR_PARAMETERS`.
    These samples are only scaled if the prior type is `PARAMETER_SCALE_...`,
    otherwise unscaled. The model has to take care of converting samples
    accordingly.
    Parameters
    ----------
    parameter_df: The PEtab parameter data frame.
    Returns
    -------
    prior: A valid pyabc.Distribution for the parameters to estimate.
    """
    # add default values
    parameter_df = petab_MS.normalize_parameter_df(parameter_df)

    prior_dct = {}

    # iterate over parameters
    for _, row in parameter_df.reset_index().iterrows():
        if row[C.ESTIMATE] == 0:
            # ignore fixed parameters
            continue

        # pyabc currently only knows objective priors, no
        #  initialization priors
        prior_type = row[C.OBJECTIVE_PRIOR_TYPE]
        # pars_str = row[C.OBJECTIVE_PRIOR_PARAMETERS]
        prior_pars = (row.lowerBound, row.upperBound)
# =======
#        pars_str = row[C.OBJECTIVE_PRIOR_PARAMETERS]
#        prior_pars = tuple(float(val) for val in pars_str.split(';'))
# >>>>>>> ece732b (add PEtab support ot FMC)
        # create random variable from table entry
        if prior_type in [C.PARAMETER_SCALE_UNIFORM,
                          C.UNIFORM]:
            lb, ub = prior_pars
            # scipy pars are location, width
            rv = pyabc.RV('uniform', loc=lb, scale=ub-lb)
        elif prior_type in [C.PARAMETER_SCALE_NORMAL,
                            C.NORMAL]:
            mean, std = prior_pars
            # scipy pars are mean, std
            rv = pyabc.RV('norm', loc=mean, scale=std)
        elif prior_type in [C.PARAMETER_SCALE_LAPLACE,
                            C.LAPLACE]:
            mean, b = prior_pars
            # scipy pars are loc=mean, scale=b
            rv = pyabc.RV('laplace', loc=mean, scale=b)
        elif prior_type == C.LOG_NORMAL:
            mean, std = prior_pars
            # petab pars are mean, std of the underlying normal distribution
            # scipy pars are s, loc, scale where s = std, scale = exp(mean)
            #  as a simple calculation shows
            rv = pyabc.RV('lognorm', s=std, loc=0, scale=np.exp(mean))
        elif prior_type == C.LOG_LAPLACE:
            mean, b = prior_pars
            # petab pars are mean, b of the underlying laplace distribution
            # scipy pars are c, loc, scale where c = 1 / b, scale = exp(mean)
            #  as a simple calculation shows
            rv = pyabc.RV('loglaplace', c=1/b, scale=np.exp(mean))
        else:
            raise ValueError(f"Cannot handle prior type {prior_type}.")

        prior_dct[row[C.PARAMETER_ID]] = rv

    # create prior distribution
    prior = pyabc.Distribution(**prior_dct)

    return prior


def get_par_map(parameter_df: pd.DataFrame) -> dict:
    """Create parameter map to be use with the Morpheus model.
    Parameters
    ----------
    parameter_df: The PEtab parameter data frame.
    Returns
    -------
    par_map: A dictionary with the xpath for all parameters.
    """
    return dict(zip(parameter_df.index, parameter_df.par_map))


def get_exp_cond_map(condition_df) -> dict:
    """get a dictionary of the experimental conditions.

    Parameters
    ----------
    condition_df: The PEtab experimental conditions data frame.
    Returns
    -------
    condition_dict: A dictionary of for the xpath and the value for each
    experimental conditions.
    """
    condition_dict = {}
    keys_values = condition_df.index.tolist()
    keys_map = condition_df.columns.tolist()[1:]
    for i, condition in enumerate(keys_values):
        tmp_dict = []
        for xpath in keys_map:
            tmp_dict.append((xpath, condition_df[xpath].tolist()[i]))
        condition_dict[condition] = dict(tmp_dict)
    return condition_dict


def get_scales(parameter_df: pd.DataFrame) -> Tuple[dict, dict]:
    """Unravel whether the priors and evaluations are on or off scale.
    Only the `parameterScale...` priors are on-scale, the other priors are on
    linear scale.
    Parameters
    ----------
    parameter_df: The PEtab parameter data frame.
    Returns
    -------
    prior_scales, scaled_scales:
        Scales for each parameter on prior and evaluation level.
    """
    # fill in objective prior columns
    parameter_df = petab_MS.normalize_parameter_df(parameter_df)
    prior_scales = {}
    scaled_scales = {}
    for _, row in parameter_df.reset_index().iterrows():
        if row[C.ESTIMATE] == 0:
            continue
        prior_scales[row[C.PARAMETER_ID]] = (
            row[C.PARAMETER_SCALE]
            if row[C.OBJECTIVE_PRIOR_TYPE] in SCALED_PRIORS
            else C.LIN)
        scaled_scales[row[C.PARAMETER_ID]] = row[C.PARAMETER_SCALE]
    return prior_scales, scaled_scales


def get_ss_post_process(observable_df: pd.DataFrame) -> Union[Callable, dict]:
    """Create a callable for the post processing function for observables.

    Parameters
    ----------
    observable_df: The PEtab observable data frame.
    Returns
    -------
    ss_post_process_dict: A Callable or dictionary of callable for each
    summary statistics.
    """
    module_path = observable_df.observableFormula.to_dict()
    post_processing_ss_dict = {}
    for key, path in module_path.items():
        head, tail = os.path.split(path)

        if head != '':
            raise ValueError(f"ss_post_process function should be within the "
                             f"Petab folder. However, now it is on: {head}.")
        filename, ext = os.path.splitext(tail)
        new_module = __import__(filename)
        post_processing_ss_dict[key] = new_module
    return post_processing_ss_dict


def import_petab_problem(
        petab_problem: petab_MS.Problem,
        model_dir) -> 'MorpheusModel':
    """
    Import model from petab problem.

    :param petab_problem:
        A petab problem containing all relevant information on the model.

    :param model_dir:
        Directory to Morpheus model.

    :return:
        The imported model.
    """

    if 'par_map' not in petab_problem.parameter_df:
        raise ValueError("'par_map' is missing in the parameter table")
    par_map = get_par_map(petab_problem.parameter_df)
    if 'parameterScale' not in petab_problem.parameter_df:
        scale = 'lin'
    else:
        scale = petab_problem.parameter_df.parameterScale.to_dict()
    ss_post_process_callable = None
    if 'observableFormula' in petab_problem.observable_df:
        ss_post_process_callable = get_ss_post_process(
            petab_problem.observable_df)
    exp_cond_map = get_exp_cond_map(petab_problem.condition_df)

    if len(exp_cond_map) < 2:
        model = MorpheusModel(model_file=model_dir,
                              par_map=par_map,
                              par_scale=scale,
                              exp_cond_map=exp_cond_map,
                              ss_post_processing=ss_post_process_callable)
        return model
    else:
        models_list = []
        models = MorpheusModels(models_list=models_list)
        for condition, val in exp_cond_map.items():
            model = MorpheusModel(model_file=model_dir,
                                  par_map=par_map,
                                  par_scale=scale,
                                  exp_cond_map={condition: val},
                                  ss_post_processing=ss_post_process_callable)
            models_list.append(model)
        return models


def get_bounds(
        parameter_df: pd.DataFrame,
        target_scale: str,
        prior_scales: dict,
        scaled_scales: dict,
        use_prior: bool) -> dict:
    """Get bounds.
    Parameters
    ----------
    parameter_df: The PEtab parameter data frame.
    target_scale:
        Scale to get the nominal parameters on.
        Can be 'lin', 'scaled', or 'prior'.
    prior_scales: Prior scales.
    scaled_scales: On-scale scales.
    use_prior: Whether to use prior overrides if of type uniform.
    Returns
    -------
    bounds: Dictionary with a (lower, upper) tuple per parameter.
    """
    parameter_df = petab_MS.normalize_parameter_df(parameter_df)

    # scale
    if target_scale == C.LIN:
        target_scales = {key: C.LIN for key in prior_scales}
    elif target_scale == 'prior':
        target_scales = prior_scales
    elif target_scale == 'scaled':
        target_scales = scaled_scales
    else:
        raise ValueError(f"Did not recognize target scale {target_scale}")

    # extract bounds
    bounds = {}
    for _, row in parameter_df.reset_index().iterrows():
        if row[C.ESTIMATE] == 0:
            # ignore fixed parameters
            continue

        key = row[C.PARAMETER_ID]

        # from lower and upper bound
        lower, upper = row[C.LOWER_BOUND], row[C.UPPER_BOUND]
        origin_scale = C.LIN

        # from prior
        prior_type = row[C.OBJECTIVE_PRIOR_TYPE]
        if use_prior and prior_type in [C.UNIFORM, C.PARAMETER_SCALE_UNIFORM]:
            pars_str = row[C.OBJECTIVE_PRIOR_PARAMETERS]
            lower, upper = tuple(float(val) for val in pars_str.split(';'))
            if prior_type == C.PARAMETER_SCALE_UNIFORM:
                origin_scale = row[C.PARAMETER_SCALE]

        # convert to target scale
        lower = rescale(
            lower, origin_scale=origin_scale, target_scale=target_scales[key])
        upper = rescale(
            upper, origin_scale=origin_scale, target_scale=target_scales[key])
        bounds[key] = (lower, upper)
    return bounds


def get_nominal_parameters(
        parameter_df: pd.DataFrame,
        target_scale: str,
        prior_scales: dict,
        scaled_scales: dict) -> pyabc.Parameter:
    """Get nominal parameters.
    Parameters
    ----------
    parameter_df: The PEtab parameter data frame.
    target_scale:
        Scale to get the nominal parameters on.
        Can be 'lin', 'scaled', or 'prior'.
    prior_scales: Prior scales.
    scaled_scales: On-scale scales.
    Returns
    -------
    par_nominal: The nominal parameters if provided in the data frame.
    """
    # unscaled parameters
    par = pyabc.Parameter(
        {key: parameter_df.loc[key, C.NOMINAL_VALUE]
         for key in prior_scales.keys()})

    # scale
    if target_scale == C.LIN:
        # nothing to be done
        return par
    elif target_scale == 'prior':
        target_scales = prior_scales
    elif target_scale == 'scaled':
        target_scales = scaled_scales
    else:
        raise ValueError(f"Did not recognize target scale {target_scale}")
    # map linear to target scales component-wise
    return map_rescale(par, origin_scales=C.LIN, target_scales=target_scales)


def rescale(val: Number, origin_scale: str, target_scale: str) -> Number:
    """Convert parameter value from origin scale to target scale.
    Parameters
    ----------
    val: Parameter value to rescale.
    origin_scale: Origin scale.
    target_scale: Target scale.
    Returns
    -------
    val: The rescaled parameter value.
    """
    if origin_scale == target_scale:
        # nothing to be done
        return val
    # origin to linear to target
    return petab_MS.scale(petab.unscale(val, origin_scale), target_scale)


def map_rescale(
        par: pyabc.Parameter,
        origin_scales: Union[dict, str], target_scales: Union[dict, str],
) -> pyabc.Parameter:
    """Rescale parameter dictionary.
    Parameters
    ----------
    par: The parameter to rescale.
    origin_scales: The origin scales.
    target_scales: The target scales.
    Returns
    -------
    par: The rescaled parameter.
    """
    # handle convenience input
    if isinstance(origin_scales, str):
        origin_scales = {key: origin_scales for key in par.keys()}
    if isinstance(target_scales, str):
        target_scales = {key: target_scales for key in par.keys()}

    # rescale each component
    for key, val in par.items():
        par[key] = rescale(
            val=val,
            origin_scale=origin_scales[key],
            target_scale=target_scales[key],
        )

    return par


def get_nr_exp_condition(condition_df):
    if condition_df.shape[0] <= 1:
        if condition_df.shape[1] <= 1:
            return "no_condition"
        else:
            return "one_condition"
    else:
        return "multiple_conditions"


# =======
#    if 'observableFormula' not in petab_problem.observable_df:
#        ss_post_process = None
#    else:
#        ss_post_process = petab_problem.observable_df.
#        observableFormula.to_dict()
#        ss_post_process_callable = get_ss_post_process(
#        petab_problem.observable_df)
#    model = MorpheusModel(model_file=model_dir,
#                          par_map=par_map,
#                          par_scale=scale,
#                          ss_post_processing=ss_post_process_callable)
#    return model
# >>>>>>> 0637f8e (allow for post_processing on simulation result comming
# from Morpheus)



"""
Summary statistics
==================

A library of summary statistics common in applications with
multi-celluar systems.

"""

from .base import (
    SumstatFun,
    IdSumstatFun,
)
from .hexagonal_cluster_sumstat import (
    ClusterCountSumstatFun,
    CCContributorsAllTpCountSumstatFun,
)
from .cell_types_cout import (
    CellCountSumstatFun,
)

__all__ = [
    'SumstatFun',
    'IdSumstatFun',
    'ClusterCountSumstatFun',
    'CCContributorsAllTpCountSumstatFun',
    'CellCountSumstatFun',
]



import pandas as pd
import numpy as np
import numbers
import os
import copy
import xml.etree.ElementTree as ET
from typing import List, Callable, Union
from types import ModuleType
import logging
import shutil
from ..sumstat import IdSumstatFun
from ..util import base as util_base
logger = logging.getLogger("FitMultiCell.Model")
try:
    from pyabc.parameters import Parameter
    from pyabc.external.base import ExternalModel
except ImportError:
    logger.error("pyabc must be installed.")


class MorpheusModel(ExternalModel):
    """
    Derived from pyabc.ExternalModel. Allows pyABC to call morpheus
    in order to do the model simulation, and then record the results
    for further processing.

    Parameters
    ----------

    model_file: str
        The XML file containing the morpheus model.
    par_map: dict
        A dictionary from str to str, the keys being the parameter ids
        to be used in pyabc, and the values xpaths in the `morpheus_file`.
    par_scale: dict or str (default: 'lin')
        A dictionary or str to state the scale used to define the parameter
        space, e.g., lin, log10, log2
    sumstat_funs: List
        List of functions to calculate summary statistics. The list entries
        are instances of fitmulticell.sumstat.SumstatFun.
    executable: str, optional
        The path to the morpheus executable. If None given,
        'morpheus' is used.
    suffix, prefix: str, optional (default: None, 'morpheus_model_')
        Suffix and prefix to use for the temporary folders created.
    dir: str, optional (default: None)
        Directory to put the temporary folders into. The default is
        the system's temporary files location. Note that these files
        are usually deleted upon system shutdown.
    show_stdout, show_stderr: bool, optional (default = False, True)
        Whether to show or hide the stdout and stderr streams.
    raise_on_error: bool, optional (default = False)
        Whether to raise on an error in the model execution, or
        just continue.
    name: str, optional (default: None)
        A name that can be used to identify the model, as it is
        saved to db. If None is passed, the model_file name is used.
    time_var: str, optional (default: "time")
        The name of the time variable as define in Morpheus model.
    ignore_list: list
        A list of columns to ignore from Morpheus output. This is introduced to
        solve the issue with result that cannot be eliminated from morpheus
        output but yet are not used in the fitting process.
    timeout: int, optional (default: np.inf)
        A time in seconds used to early stop Morpheus simulation when exceed.
    ss_post_processing: Union[Callable, dict], optional
        A callable function to perform post processing on Morpheus output. If
        a dict is passed, then specific function will be applied to each
        summary statistics.
    output_file: str, optional (default: "logger.csv")
        A name of the file containing the simulation output.
    """

    def __init__(self,
                 model_file: str,
                 par_map: dict,
                 par_scale: Union[dict, str] = "lin",
                 exp_cond_map: dict = None,
                 sumstat_funs: List = None,
                 executable: str = "morpheus",
                 suffix: str = None,
                 prefix: str = "morpheus_model_",
                 dir: str = None,
                 clean_simulation: bool = False,
                 show_stdout: bool = False,
                 show_stderr: bool = True,
                 raise_on_error: bool = False,
                 name: str = None,
                 time_var: str = "time",
                 ignore_list: list = [],
                 timeout: int = np.inf,
                 ss_post_processing: Union[Callable, dict] = None,
                 output_file: str = "logger.csv"
                 ):
        if name is None:
            name = model_file
        super().__init__(
            executable=executable,
            file=model_file,
            fixed_args=None,
            create_folder=True,
            suffix=suffix, prefix=prefix, dir=dir,
            clean_simulation=clean_simulation,
            show_stdout=show_stdout,
            show_stderr=show_stderr,
            raise_on_error=raise_on_error,
            name=name)
        self.par_map = par_map
        self.par_scale = par_scale
        self.exp_cond_map = exp_cond_map
        if sumstat_funs is None:
            if self.exp_cond_map is None:
                sumstat_funs = [IdSumstatFun()]
            else:
                sumstat_funs = [IdSumstatFun(
                    name=list(self.exp_cond_map.keys())[0])]
        self.clean_simulation = clean_simulation
        self.sumstat_funs = sumstat_funs
        self._check_sumstat_funs()
        self.time_var = time_var
        self.ignore_list = ignore_list
        self.timeout = timeout
        self.ss_post_processing = ss_post_processing
        self.output_file = output_file
        self._sanity_chceck_model()

    def __str__(self):
        s = f"MorpheusModel {{\n" \
            f"\tname      : {self.name}\n" \
            f"}}"
        return s

    def __repr__(self):
        return self.__str__()

    def __call__(self, pars: Parameter):
        """
        This function is used in ABCSMC (or rather the sample() function,
        which redirects here) to simulate data for given parameters `pars`.
        """

        # TODO: move all the content to another function and just call here.

        # create target on file system
        loc = self.eh.create_loc()
        file_ = os.path.join(loc, "model.xml")

        # write new file with parameter modifications
        self.write_modified_model_file(file_, pars)

        # create command
        cmd = self.eh.create_executable(loc)
        self.eh.timeout = self.timeout
        cmd = cmd + f" -file={file_} -outdir={loc}"

        # call the model
        status = self.eh.run(args=pars, cmd=cmd, loc=loc)

        # compute summary statistics
        if status == -15:
            msg = (f"Simulation exceed time limit limit: {self.timeout}s "
                   f"for arguments {pars}: returncode {status}.")
            logger.warning(msg)
            return -15
        sumstats = self.compute_sumstats(loc)

        # remove simulation output
        if self.clean_simulation:
            clean_simulation_output(loc)

        # perform data post-process on Morpheus output
        sumstats = self.call_post_processing_ss(sumstats)
        return sumstats

    def get_parmap_xpath_attr(self, key, attrib='value'):
        """
        Get the xpath and for the parameter of interest

        Parameters
        ----------
        key: str
            name of parameter of interest.
        attrib: str
            the type of attribute that need to be changed on the xml file.

        """

        par = self.par_map[key]
        if isinstance(par, str):
            return par, attrib
        elif isinstance(par, (list, tuple)) and len(par) == 2:
            return par[0], par[1]
        else:
            raise TypeError(
                f"par_map[{key}] should be a str or a list/tuple of length 2")

    def get_expcondmap_xpath_attr(self, key, attrib='value'):
        """
        Get the xpath and for the experimental conditions of interest

        Parameters
        ----------
        key: str
            name of experimental condition of interest.
        attrib: str
            the type of attribute that need to be changed on the xml file.

        """

        exp_cond = self.exp_cond_map[key]
        if isinstance(exp_cond, str):
            return exp_cond, attrib
        elif isinstance(exp_cond, (list, tuple)) and len(exp_cond) == 2:
            return exp_cond[0], exp_cond[1]
        else:
            raise TypeError(
                f"par_map[{key}] should be a str or a list/tuple of length 2")

    def write_modified_model_file(self, file_, pars):
        """
        Write a modified version of the morpheus xml file to the target
        directory.
        """
        temp_par = copy.deepcopy(pars)
        rescaled_pars = util_base.scaling_parameter(temp_par, self.par_scale)
        tree = ET.parse(self.eh.file)
        root = tree.getroot()
        for key, val in rescaled_pars.items():
            xpath, attr = self.get_parmap_xpath_attr(key)
            node = root.findall(xpath)
            if node.__len__() == 1:
                node[0].set(attr, str(val))
            else:
                raise KeyError(
                    f"Key {key} is not unique or does not exist.")

        # TODO: better handle the attr for exp_cond_map
        attr = 'value'
        if self.exp_cond_map:
            if len(self.exp_cond_map) < 2:
                for condition, val in self.exp_cond_map.items():
                    for element, inner_val in val.items():
                        if type(inner_val) != list:
                            xpath = element
                        else:
                            xpath = element
                            attr = inner_val[0]
                            inner_val = inner_val[1]
                        node = root.findall(xpath)
                        if node.__len__() == 1:
                            node[0].set(attr, str(inner_val))
                        else:
                            raise KeyError(
                                f"condition {condition} "
                                f"is not unique or does not exist.")
            else:
                raise KeyError(
                    "It seems that the model has more that one condition. "
                    "Please try to use only one condition or"
                    " use appropriate model class.")

        tree.write(file_, encoding="utf-8", xml_declaration=True)

    def compute_sumstats(self, loc):
        """
        Compute summary statistics from the simulated data according to the
        provided list of summary statistics functions.
        """
        sumstat_dict = {'loc': loc}
        if "IdSumstat" in self.sumstat_funs[0].name:
            sumstat_dict = self.sumstat_funs[0](
                loc, self.ignore_list, self.output_file)
            sumstat_dict['loc'] = loc

            # here add the prepare function,e.g., 0: val,val,val
            # safe_append_sumstat(sumstat_dict, sumstat, self.sumstat_funs[0])
            return sumstat_dict
        else:
            for sumstat_fun in self.sumstat_funs:
                sumstat = sumstat_fun(loc, self.ignore_list, self.output_file)
                # here add the prepare function,e.g., 0: val,val,val
                safe_append_sumstat(sumstat_dict, sumstat, sumstat_fun.name)
            return sumstat_dict

    def _check_sumstat_funs(self):
        """
        Check sumstat functions for validity.
        """
        names = [ssf.name for ssf in self.sumstat_funs]
        if not len(set(names)) == len(names):
            raise AssertionError(
                f"The summary statistics passed to MorpheusModel must have"
                f"unique names, but obtained {names}")

    def _sanity_chceck_model(self):
        """
        Sanity check of the model upon construction.
        """

        loc = self.eh.create_loc()
        file_ = os.path.join(loc, "model.xml")

        # write new file with for the model
        tree = ET.parse(self.eh.file)
        tree.write(file_, encoding="utf-8", xml_declaration=True)

        # create command
        cmd = self.eh.create_executable(loc)
        # self.eh.timeout = self.timeout
        cmd = cmd + f" -file={file_} -outdir={loc}"

        # call the model
        status = self.eh.run(cmd=cmd, loc=loc)
        if status["returncode"] == 0:
            logger.info("Successfully loaded model")
        else:
            raise ValueError(
                f"The model is not running well. "
                f"The simulation seems to fail. "
                f"Returncode: {status['returncode']}")
        clean_simulation_output(loc)

    def call_post_processing_ss(self, sumstats, function_name='main'):
        if self.ss_post_processing is not None:
            if isinstance(self.ss_post_processing, Callable):
                sumstats = self.ss_post_processing(sumstats)
            elif isinstance(self.ss_post_processing, dict):
                if not set(self.ss_post_processing.keys()).issubset(
                        sumstats.keys()):
                    raise ValueError(
                        "the keys on the 'ss_post_processing' does not "
                        "match the one on the summary statistics names.")
                if isinstance(list(self.ss_post_processing.values())[0],
                              Callable):
                    sumstats = self._call_post_processing_ss_use_function(
                        sumstats)
                elif isinstance(list(self.ss_post_processing.values())[0],
                                ModuleType):
                    sumstats = self._call_post_processing_ss_use_module(
                        sumstats, function_name)
            else:
                raise ValueError(
                    f"the type of 'post_processing_ss' should be str or dict."
                    f" However, {type(self.ss_post_processing)} was given.")
        return sumstats

    def _call_post_processing_ss_use_module(self, sumstats, function_name):
        sumstat_pp = {}
        for key, module in self.ss_post_processing.items():
            try:
                # sumstat_pp[key] = module.main(sumstats)
                func = getattr(self.ss_post_processing[key],
                               function_name, None)
                sumstat_pp[key] = func(sumstats[key])
            except Exception as e:
                raise RuntimeError(
                    f"the selected ss_post_processing function "
                    f"can not be called. Be sure that the main "
                    f"function called `main(). `{e}")
        return sumstat_pp

    def _call_post_processing_ss_use_function(self, sumstats):
        sumstat_pp = {}
        for key, function in self.ss_post_processing.items():
            try:
                sumstat_pp[key] = function(sumstats[key])
            except Exception as e:
                raise RuntimeError(
                    f"the selected ss_post_processing function "
                    f"can not be called. `{e}")
        return sumstat_pp


class MorpheusModels(ExternalModel):
    """
    Derived from pyabc.ExternalModel. Allows pyABC to call morpheus
    in order to do the models simulation, and then record the results
    for further processing.

    Parameters
    ----------
    model_file: str
        The XML file containing the morpheus model.
    par_map: dict
        A dictionary from str to str, the keys being the parameter ids
        to be used in pyabc, and the values xpaths in the `morpheus_file`.
    sumstat_funs: List
        List of functions to calculate summary statistics. The list entries
        are instances of fitmulticell.sumstat.SumstatFun.
    executable: str, optional
        The path to the morpheus executable. If None given,
        'morpheus' is used.
    suffix, prefix: str, optional (default: None, 'morpheus_model_')
        Suffix and prefix to use for the temporary folders created.
    dir: str, optional (default: None)
        Directory to put the temporary folders into. The default is
        the system's temporary files location. Note that these files
        are usually deleted upon system shutdown.
    show_stdout, show_stderr: bool, optional (default = False, True)
        Whether to show or hide the stdout and stderr streams.
    raise_on_error: bool, optional (default = False)
        Whether to raise on an error in the model execution, or
        just continue.
    name: str, optional (default: None)
        A name that can be used to identify the model, as it is
        saved to db. If None is passed, the model_file name is used.
    time_var: str, optinal (default: "time")
        The name of the time variable as define in Morpheus model.
    ignore_list: list
        A list of columns to ignore from Morpheus output. This is introduced to
        solve the issue with result that cannot be eliminated from morpheus
        output but yet are not used in the fitting process.
    timeout: int, optional (default: np.inf)
        A time in seconds used to early stop Morpheus simulation when exceed.
    ss_post_processing: Union[Callable, dict], optional
        A callable function to perform post processing on Morpheus output. If
        a dict is passed, then specific function will be applied to each
        summary statistics.
    """

    def __init__(self,
                 models_list: List[MorpheusModel],
                 name: str = None
                 ):
        self.models_list = models_list
        self.name = name

    def __str__(self):
        s = f"MorpheusModels {{\n" \
            f"\tname      : {self.name}\n" \
            f"}}"
        return s

    def __repr__(self):
        return self.__str__()

    def __call__(self, pars: Parameter):
        """
        This function is used in ABCSMC (or rather the sample() function,
        which redirects here) to simulate data for given parameters `pars`
        and given experimental conditions.
        """

        # TODO: move all the content to another function and just call here.

        # a list that will hold result for each experimental condition.
        sumstats_all = {}
        # create target on file system
        for i, model in enumerate(self.models_list):
            # cond_sumstats = {}
            sumstats = model(pars)
            # for ss_key, ss_val in sumstats.items():
            #     cond_sumstats[list(model.exp_cond_map.keys())[0]
            #     + '__' + ss_key] = sumstats[ss_key]
            sumstats_all.update(sumstats)
        return sumstats_all

    def get_parmap_xpath_attr(self, key, attrib='value'):
        """
        Get the xpath and for the parameter of interest

        Parameters
        ----------
        key: str
            name of parameter of interest.
        attrib: str
            the type of attribute that need to be changed on the xml file.

        """
        # TODO: this function is written twice
        par = self.par_map[key]
        if isinstance(par, str):
            return par, attrib
        elif isinstance(par, (list, tuple)) and len(par) == 2:
            return par[0], par[1]
        else:
            raise TypeError(
                f"par_map[{key}] should be a str or a list/tuple of length 2")

    def write_modified_models_file(self, file_, pars, exp_cod):
        """
        Write a modified version of the morpheus xml file to the target
        directory.
        """
        temp_par = copy.deepcopy(pars)
        rescaled_pars = util_base.scaling_parameter(temp_par, self.par_scale)
        tree = ET.parse(self.eh.file)
        root = tree.getroot()
        # parameters
        for key, val in rescaled_pars.items():
            xpath, attr = self.get_parmap_xpath_attr(key)
            node = root.findall(xpath)
            if node.__len__() == 1:
                node[0].set(attr, str(val))
            else:
                raise KeyError(
                    f"Key {key} is not unique or does not exist.")
        # conditions
        for key, val in exp_cod.items():
            xpath, attr = self.get_parmap_xpath_attr(key)
            node = root.findall(xpath)
            if node.__len__() == 1:
                node[0].set(attr, str(val))
            else:
                raise KeyError(
                    f"Key {key} is not unique or does not exist.")

        tree.write(file_, encoding="utf-8", xml_declaration=True)

    def call_post_processing_ss(self, sumstats, function_name='main'):
        if self.ss_post_processing is not None:
            if isinstance(self.ss_post_processing, Callable):
                sumstats = self.ss_post_processing(sumstats)
            elif isinstance(self.ss_post_processing, dict):
                if not set(self.ss_post_processing.keys()).issubset(
                        sumstats.keys()):
                    raise ValueError(
                        "the keys on the 'ss_post_processing' does not "
                        "match the one on the summary statistics names.")
                if isinstance(list(self.ss_post_processing.values())[0],
                              Callable):
                    sumstats = self._call_post_processing_ss_use_function(
                        sumstats)
                elif isinstance(list(self.ss_post_processing.values())[0],
                                ModuleType):
                    sumstats = self._call_post_processing_ss_use_module(
                        sumstats, function_name)
            else:
                raise ValueError(
                    f"the type of 'post_processing_ss' should be str or dict."
                    f" However, {type(self.ss_post_processing)} was given.")
        return sumstats

    def _call_post_processing_ss_use_module(self, sumstats, function_name):
        sumstat_pp = {}
        for key, module in self.ss_post_processing.items():
            try:
                # sumstat_pp[key] = module.main(sumstats)
                func = getattr(self.ss_post_processing[key],
                               function_name, None)
                sumstat_pp[key] = func(sumstats[key])
            except Exception as e:
                raise RuntimeError(
                    f"the selected ss_post_processing function "
                    f"can not be called. Be sure that the main "
                    f"function called `main(). `{e}")
        return sumstat_pp

    def _call_post_processing_ss_use_function(self, sumstats):
        sumstat_pp = {}
        for key, function in self.ss_post_processing.items():
            try:
                sumstat_pp[key] = function(sumstats[key])
            except Exception as e:
                raise RuntimeError(
                    f"the selected ss_post_processing function "
                    f"can not be called. `{e}")
        return sumstat_pp


def safe_append_sumstat(sumstat_dict, sumstat, key):
    types_ = (numbers.Number, np.ndarray, pd.DataFrame)
    if isinstance(sumstat, types_):
        if key in sumstat_dict:
            raise KeyError(
                f"Key {key} for sumstat {sumstat} already in the "
                f"sumstat dict {sumstat_dict}.")
        sumstat_dict[key] = sumstat
        return
    if isinstance(sumstat, dict):
        if key in sumstat_dict:
            raise KeyError(
                f"Key {key} for sumstat {sumstat} already in the "
                f"sumstat dict {sumstat_dict}.")
        sumstat_dict.update(sumstat)
        return
    raise ValueError(
        f"Type {type(sumstat)} of sumstat {sumstat} "
        f"is not permitted.")


def clean_simulation_output(loc):
    """
    Remove the simulation output directory after calculating the
    summary statistics.

    Parameters
    ----------
    loc: str
        Location of the simulation directory.
    """

    shutil.rmtree(loc, ignore_errors=True)



import numpy as np
from scipy import stats
from typing import List
import logging

logger = logging.getLogger(__name__)

try:
    from pyabc.distance import Distance
except ImportError:
    logger.error("pyabc must be installed.")


class KolmogorovSmirnovDistance(Distance):
    """
    Compute a 2 sample Kolmogorov Smirnov distance from the
    simulated and observed data.
    """

    def __init__(self, keys: List = None):
        self.keys = keys

    def __call__(self, x: dict, x_0: dict, t: int = None, par: dict = None):
        if self.keys is not None:
            keys = self.keys
        else:
            keys = list(x_0.keys())

        sim_vals = []
        obs_vals = []
        for key in keys:
            sim_vals.append(x[key])
            obs_vals.append(x_0[key])

        sim_vals = np.array(sim_vals).flatten()
        obs_vals = np.array(obs_vals).flatten()

        res = stats.ks_2samp(sim_vals, obs_vals)
        return res.statistic



import pyabc
import fitmulticell as fmc
import os
import numpy as np
import tempfile


def test_nothing():
    pass


def _test_mapk():
    """
    Just runs an analysis with the MAPK model to ensure that at least
    not execution errors occur.
    """
    file_ = "doc/example/MAPK_SBML.xml"
    par_map = {'V1': './CellTypes/CellType/System/Constant[@symbol="V1"]',
               'V2': './CellTypes/CellType/System/Constant[@symbol="V2"]',
               'k3': './CellTypes/CellType/System/Constant[@symbol="k3"]'}

    model = fmc.model.MorpheusModel(
        file_, par_map=par_map,
        # executable="ch-run ../../test/morpheus_docker -- /usr/bin/morpheus",
        show_stdout=False, show_stderr=True,
        raise_on_error=False)

    true_pars = {'V1': 2.7, 'V2': 0.25, 'k3': 0.025}
    limits = {key: (0.5 * val, 2 * val) for key, val in true_pars.items()}

    # generate data
    observed_data = model.sample(true_pars)

    prior = pyabc.Distribution(**{key: pyabc.RV("uniform", lb, ub - lb)
                                  for key, (lb, ub) in limits.items()})

    def distance(val1, val2):
        d = np.sum([np.sum(np.abs(val1[key] - val2[key]))
                    for key in ['IdSumstat__MAPK_P', 'IdSumstat__MKK_P']])
        return d

    abc = pyabc.ABCSMC(model, prior, distance, population_size=5)
    db_path = "sqlite:///" + os.path.join(tempfile.gettempdir(), "test.db")
    abc.new(db_path, observed_data)

    # run
    h = abc.run(max_nr_populations=3)

    pyabc.visualization.plot_epsilons(h)
    df, w = h.get_distribution(t=h.max_t)
    pyabc.visualization.plot_kde_matrix(
        df, w, limits=limits, refval=true_pars)



import numpy as np
import math
from typing import List

from .base import SumstatFun
from ..util import tsv_to_df


class ClusterCountSumstatFun(SumstatFun):
    """
    Total number of clusters for all time points.
    """

    def __init__(self,
                 field_of_interest: str,
                 cluster_cell_type: List,
                 time_step: int = None,
                 time_interval: List = None,
                 time_symbol: str = 't',
                 name: str = "ClusterCount",
                 ):
        """
        Parameters
        ----------
        field_of_interest: str
            The field of interest that contains the clusters data.
        cluster_cell_type: list
            List of cell type that are member of the cluster.
        time_step: int (optional)
            The time step between time points. If not specified,
            the time step of the Morpheus file will be used.
        time_interval: list [start_time_point, end_time_point] (optional)
            You can use the time_interval to define the starting and ending
            point that is in the range of the origianl time interval of the
            simulation.
            If not specified, the original time interval will be used.
        time_symbol: str (default=t)
            The time symbol, as specified in the morpheus model.
        """
        super().__init__(name)
        self.field_of_interest = field_of_interest
        self.cluster_cell_type = cluster_cell_type
        self.time_step = time_step
        self.time_interval = time_interval
        self.time_symbol = time_symbol

    def __call__(self, loc):
        """
        Parameters
        ----------
        loc: str
            The simulation folder.

        Returns
        -------
        cluster_count: int
            The total number of clusters for all time points.
        """
        df = tsv_to_df(loc, "logger.csv")
        return get_clusters_count(
            df, self.field_of_interest, self.cluster_cell_type,
            self.time_step, self.time_interval, self.time_symbol)


def get_clusters_count(
        logger_df, field_of_interest,
        cluster_cell_type, time_step=None, time_interval=None,
        time_symbol='t'):
    """
    Get the total number of clusters for all time points.
    """
    # check if time_step is specified, if not then calculate its value from
    # the distance between two time points from the morpheus logger file
    if not time_step:
        time_steps = logger_df[time_symbol].unique()
        time_step = time_steps[1] - time_steps[0]
    # starting and ending time point
    if time_interval is None:
        start = logger_df[time_symbol].unique().tolist()[0]
        end = logger_df[time_symbol].unique().tolist()[-1]
    else:
        start = time_interval[0]
        end = time_interval[1]

    cluster_per_time_point = {}
    # filter the data to only consider a specifice time point
    filtered_data = logger_df[logger_df[time_symbol] == start]
    number_of_cells = filtered_data.shape[0]
    # calculate the edge size of the hexagonal grid
    edge = math.ceil(math.sqrt((number_of_cells - 1) / 3))
    # iterate through all time points
    for j in range(start, end + 1, time_step):
        # filter the data to only consider a specific time point
        filtered_data = logger_df[logger_df[time_symbol] == j]
        # only consider the data on the field of interest
        morpheus_logger_list = filtered_data[field_of_interest].tolist()
        cluster_count = 0
        full_list_of_cluster = []
        # iterate through all cells in the specified time point
        for index in range(len(morpheus_logger_list)):
            # check if the cell belongs to the cluster and it is not already
            # considered
            if (morpheus_logger_list[index] in cluster_cell_type) and\
                    (index not in full_list_of_cluster):
                index_list_of_cluster = []
                # find the neighbor cells list for a specific cell
                neighbor_cells_list = find_neighbor_list(index, edge)
                index_list_of_cluster.append(index)
                # check if the neighbor cells also belong to the same cluster
                index_list_of_cluster = check_for_connecting_cluster_loop(
                    cluster_cell_type, edge, neighbor_cells_list,
                    index_list_of_cluster, morpheus_logger_list)
                full_list_of_cluster = full_list_of_cluster + \
                    index_list_of_cluster
                cluster_count = cluster_count + 1
            morpheus_logger_list_unique = sorted(
                list(set(morpheus_logger_list)))
        if morpheus_logger_list_unique == sorted(cluster_cell_type):
            cluster_count = 1
        cluster_per_time_point[j] = cluster_count

    return {"cluster_count": list(cluster_per_time_point.values())}


def find_offset_list(edge: int) -> list:
    """
    Get the offset for all rows to be used to identify the neighbor cells.

    Parameters
    ----------
    edge: int
        The edge size of the hexagonal grid.

    Returns
    -------
    offset: list
        A list of offsets for the entire hexagonal grids.
    """
    k = 0
    counter_index = 0
    offset = []
    for i in range(0, edge * 2 - 2):
        if i < edge - 1:
            k = edge + i
            offset.append(k)
        else:
            offset.append(k - counter_index)
            counter_index = counter_index + 1
    return offset


def find_offset_int(edge: int, row: int) -> int:
    """
    Get the offset of the neighbor cells.

    Parameters
    ----------
    edge: int
        The edge size of the hexagonal grid.
    row: int
        The number of which we want to find its offset.

    Returns
    -------
    offset: int
        The offset of the specific row.
    """
    k = 0
    counter_index = 0
    offset = []
    for i in range(0, edge * 2 - 2):
        if i < edge - 1:
            k = edge + i
            offset.append(k)
        else:
            offset.append(k - counter_index)
            counter_index = counter_index + 1
    return offset[row]


def find_neighbor_list(cell_id, edge: int, nn=6):
    """
    Get the list of neighbor cells for a specific cell.

    Parameters
    ----------
    cell_id: int
        The ID of the cell that we want to find its neighbor.
    edge: int
        The edge size of the hexagonal grid.
    nn: int, optional (default=6)
        Number of neighbor cells. E.g. a hexagonal
        cell might have nn=6 because it has 6 neighbors.

    Returns
    -------
    list_of_neighbor: list
        List of neighbor cells.
    """
    # the current function works only when the number of neighbor cells is 5
    if nn != 6:
        return "Currently this function can only deal with neighbors = 6!"
    edge_cells_list = find_edge_cell(edge)
    # check if cell_id is out of the range of the hexagonal grid of the model
    if cell_id < edge_cells_list[0][0] or\
            cell_id > edge_cells_list[1][edge * 2 - 2]:
        print("the cell_id is out of the range")
        return -1
    row = find_row_of_cell(cell_id, edge)
    offset = find_offset_list(edge)
    list_of_neighbor = []
    # deals with special scenarios, e.g., if cell is on the first row,
    # it will have only 4 neighbors
    # check if cell is in the last row of hexagon
    if row == edge * 2 - 2:
        list_of_neighbor.append(cell_id - offset[row - 1])
        list_of_neighbor.append(cell_id - offset[row - 1] - 1)
        if cell_id == edge_cells_list[1][edge * 2 - 2] - edge + 1:
            list_of_neighbor.append(cell_id + 1)
        elif cell_id == edge_cells_list[1][edge * 2 - 2]:
            list_of_neighbor.append(cell_id - 1)
        else:
            list_of_neighbor.append(cell_id + 1)
            list_of_neighbor.append(cell_id - 1)
    # check if cell is in the first row of hexagon
    elif row == edge_cells_list[0][0]:
        list_of_neighbor.append(cell_id + offset[row])
        list_of_neighbor.append(cell_id + offset[row] + 1)
        if cell_id == edge_cells_list[0][0]:
            list_of_neighbor.append(cell_id + 1)
        elif cell_id == edge - 1:
            list_of_neighbor.append(cell_id - 1)
        else:
            list_of_neighbor.append(cell_id + 1)
            list_of_neighbor.append(cell_id - 1)
    # check if cell not on the edge of hexagon
    elif cell_id not in edge_cells_list:
        list_of_neighbor.append(cell_id - 1)
        list_of_neighbor.append(cell_id + 1)
        list_of_neighbor.append(cell_id + offset[row])
        list_of_neighbor.append(cell_id + offset[row] + 1)
        list_of_neighbor.append(cell_id - offset[row - 1])
        list_of_neighbor.append(cell_id - offset[row - 1] - 1)

    # check if cell not on left edge of hexagon
    elif cell_id in edge_cells_list[0][:]:
        list_of_neighbor.append(cell_id + 1)
        list_of_neighbor.append(cell_id + offset[row] + 1)
        list_of_neighbor.append(cell_id - offset[row - 1])
        if edge < row + 1:
            list_of_neighbor.append(cell_id - offset[row - 1] - 1)
        if edge > row + 1:
            list_of_neighbor.append(cell_id + offset[row])
    # check if cell not on right edge of hexagon
    elif cell_id in edge_cells_list[1][:]:
        list_of_neighbor.append(cell_id - 1)
        list_of_neighbor.append(cell_id + offset[row])
        list_of_neighbor.append(cell_id - offset[row - 1] - 1)
        if edge < row + 1:
            list_of_neighbor.append(cell_id - offset[row - 1])
        if edge > row + 1:
            list_of_neighbor.append(cell_id + offset[row] + 1)
    return list_of_neighbor


def find_edge_cell(edge: int):
    """
    Get the list cells that are in the edge of the hexagonal grid.

    Parameters
    ----------
    edge: int
        The edge size of the hexagonal grid.

    Returns
    -------
    edge_cell_list: list
        List of edge cells.
    """
    edge_cell_list = np.zeros((2, edge * 2 - 1))
    k = edge
    temp = 0
    # iterate through all rows in the grid and locate edge cells for each row
    for i in range(0, edge * 2 - 1):
        if i == 0:
            edge_cell_list[0, i] = temp
        elif i < edge:
            temp = temp + k
            edge_cell_list[0, i] = temp
            k = k + 1
        else:
            temp = temp + k
            edge_cell_list[0, i] = temp
            k = k - 1
    k = edge + 1
    temp = edge - 1
    for i in range(0, edge * 2 - 1):
        if i == 0:
            edge_cell_list[1, i] = temp
        elif i < edge:
            temp = temp + k
            edge_cell_list[1, i] = temp
            k = k + 1
        elif i == edge:
            k = k - 2
            temp = temp + k
            edge_cell_list[1, i] = temp
        else:
            k = k - 1
            temp = temp + k
            edge_cell_list[1, i] = temp
    edge_cell_list = edge_cell_list.astype(int)
    return edge_cell_list


def check_neighbor_inactive_infected_cell(
        morpheus_neighbor_list, index, inactive_cell, cluster_cell_type):
    """
    Check if the cell (index) is infected or not.

    Parameters
    ----------
    morpheus_neighbor_list: list
        The field of interest in the structure of list.
    index: int
        The cell id of the cell that will be checked.
    inactive_cell: list
        The list of the inactive infected cell.
    cluster_cell_type: list
        List of cell type that are member of the cluster.

    Returns
    -------
    inactive_cell: int
        The cumulative list of the inactive infected cells.
    """
    if len(morpheus_neighbor_list) > 0 and \
            all(elem in cluster_cell_type for elem in morpheus_neighbor_list):
        inactive_cell.append(index + 1)
    return inactive_cell


def find_row_of_cell(cell_id: int, edge: int):
    """
    Get the list of neighbor cells for a specific cell.

    Parameters
    ----------
    cell_id: int
        The ID of the cell whose row we want to find.
    edge: int
        The edge size of the hexagonal grid.

    Returns
    -------
    row_number: int
        the row number of the cell with (cell_id)
    """
    edge_cell_list = find_edge_cell(edge)
    row_number = 0
    for i in range(0, len(edge_cell_list[0])):
        if edge_cell_list[0][i] <= cell_id <= edge_cell_list[1][i]:
            row_number = i
            break
    return row_number


def check_for_connecting_cluster_loop(
        cluster_cell_type, edge, neighbor_cells_list,
        index_list_of_cluster, morpheus_logger_list) -> list:
    """
    Check whether the neighbor_cell_list ist member of cluster.

    Parameters
    ----------
    edge: int
        The edge size of the hexagonal grid.
    neighbor_cells_list: list
        A list for the neighbor cell.
    index_list_of_cluster: list
        The list of cluster member.
    morpheus_logger_list: pd.Series
        A pandas time series for the field of interest taken from the
        morpheus logger file.

    Returns
    -------
    index_list_of_cluster: list
        the list of cluster member
    """
    # iterate through all the cells in the neighbor_cells_list
    # check if they are connected to the cluster or not
    for neighbor_cell in neighbor_cells_list:
        # check if the cell has the same characteristics as the rest
        # of the cluster
        if morpheus_logger_list[neighbor_cell] in cluster_cell_type:
            if neighbor_cell not in index_list_of_cluster:
                index_list_of_cluster.append(neighbor_cell)
            new_neighbor_cells_list = find_neighbor_list(
                neighbor_cell, edge)
            # check if the cell is still not discovered as part of the
            # cluster in the current iteration
            for cell in new_neighbor_cells_list:
                if cell not in neighbor_cells_list:
                    neighbor_cells_list.append(cell)
    return index_list_of_cluster


class ClusterSizeSumstatFun(SumstatFun):
    """
    Total number of clusters and size of each cluster.
    """

    def __init__(
            self,
            field_of_interest: str,
            cluster_cell_type: List,
            time_point: int,
            time_symbol: str = 't',
            name: str = "ClusterSize"):
        """
        Parameters
        ----------
        logger_df: pd.DataFrame
            The logger file in pandas dataframe format.
        field_of_interest: str
            The field of interest that contains the clusters data.
        cluster_cell_type: list
            A list of cell type that are member of the cluster.
        time_point: int
            The time point where we want count the number of clusters.
        time_symbol: str (default=t)
            The time symbol, as specified in the morpheus model.
        """
        super().__init__(name)
        self.field_of_interest = field_of_interest
        self.cluster_cell_type = cluster_cell_type
        self.time_point = time_point
        self.time_symbol = time_symbol

    def __call__(self, loc):
        """
        Returns
        -------
        cluster_count: dict
            A dictionary that contains the index and the member of each
            cluster.
        """
        df = tsv_to_df(loc, "logger.csv")
        return get_clusters_sizes_tp(
            df, self.field_of_interest, self.cluster_cell_type,
            self.time_point, self.time_symbol)


def get_clusters_sizes_tp(
        logger_df, field_of_interest, cluster_cell_type,
        time_point, time_symbol='t'):
    """
    Get the total number of clusters and the size of each cluster
    for a specific time point.
    """
    full_list_of_cluster = []
    dict_cluster_member = {}
    # filter the data to only consider a specific time point
    filtered_data = logger_df[logger_df[time_symbol] == time_point]
    number_of_cells = filtered_data.shape[0]
    # calculate the edge size of the hexagonal grid
    edge = math.ceil(math.sqrt((number_of_cells - 1) / 3))
    morpheus_logger_list = filtered_data[field_of_interest].tolist()
    # iterate through al cells in the specified time point
    for index in range(len(morpheus_logger_list)):
        # check if the cell belongs to the cluster and it is not already
        # considered
        if (morpheus_logger_list[index] in cluster_cell_type) and\
                (index + 1 not in full_list_of_cluster):
            index_list_of_cluster = []
            # find the neighbor cells list for a specific cell
            neighbor_cells_list = find_neighbor_list(index, edge)
            index_list_of_cluster.append(index)
            # check if the neighbor cells also belong to the same cluster
            index_list_of_cluster = check_for_connecting_cluster_loop(
                cluster_cell_type, edge, neighbor_cells_list,
                index_list_of_cluster, morpheus_logger_list)
            index_list_of_cluster = [x + 1 for x in index_list_of_cluster]
            dict_cluster_member[index + 1] = len(index_list_of_cluster)
            full_list_of_cluster = full_list_of_cluster + index_list_of_cluster
    return dict_cluster_member


class CCNonContributorsTpSumstatFun(SumstatFun):
    """
    List of inactive cell in the grid at a specific time point.
    """

    def __init__(
            self,
            field_of_interest: str,
            cluster_cell_type: List,
            time_point: int,
            time_symbol: str = 't',
            name: str = "CCNonContributorsTp"):
        """
        Parameters
        ----------
        logger_df: pd.DataFrame
            The logger file in pandas dataframe format.
        field_of_interest: str
            The field of interest that contains the clusters data
        cluster_cell_type: list
            A list of cell type that are member of the cluster
        time_point: int
            The time point where we want count the number of clusters
        time_symbol: str (default=t)
            The time symbol, as specified in the morpheus model
        """
        super().__init__(name)
        self.field_of_interest = field_of_interest
        self.cluster_cell_type = cluster_cell_type
        self.time_point = time_point
        self.time_symbol = time_symbol

    def __call__(self, loc):
        """
        Returns
        -------
        inactive_cell_list: list
            A list of all inactive infected cell, that means all cell that does
            not contribute to the cell-to-cell infection.
        """
        df = tsv_to_df(loc, "logger.csv")
        return get_cc_non_contributors_tp(
            df, self.field_of_interest, self.cluster_cell_type,
            self.time_point, self.time_symbol)


def get_cc_non_contributors_tp(
        logger_df, field_of_interest, cluster_cell_type,
        time_point, time_symbol='t'):
    """
    Get the list of inactive cell in the grid at a specific time point.
    """
    # filter the data to only consider a specific time point
    filtered_data = logger_df[logger_df[time_symbol] == time_point]
    number_of_cells = filtered_data.shape[0]
    # calculate the edge size of the hexagonal grid
    edge = math.ceil(math.sqrt((number_of_cells - 1) / 3))
    morpheus_logger_list = filtered_data[field_of_interest].tolist()
    inactive_cell_list = []
    full_list_of_cluster = []
    # iterate through all cells in the specified time point
    for index in range(len(morpheus_logger_list)):
        if (morpheus_logger_list[index] in cluster_cell_type) and \
                (index + 1 not in full_list_of_cluster):
            index_list_of_cluster = []
            neighbor_cells_list = find_neighbor_list(index, edge)
            morpheus_neighbor_list = []
            # iterate through all neighbor cells
            for i in neighbor_cells_list:
                morpheus_neighbor_list.append(morpheus_logger_list[i])
            index_list_of_cluster.append(index)
            # check if neighbor cells are CC active or not
            inactive_cell_list = check_neighbor_inactive_infected_cell(
                morpheus_neighbor_list, index, inactive_cell_list,
                cluster_cell_type)
    return inactive_cell_list


class CCContributorsAllTpCountSumstatFun(SumstatFun):
    """
    Total number of active infected cells in the grid at a
    specific time point.
    """

    def __init__(
            self,
            field_of_interest: str,
            cluster_cell_type: List,
            time_step: int = None,
            time_interval: List = None,
            time_symbol: str = 't',
            name: str = "CCContributorsAllTpCount"):
        """
        Parameters
        ----------
        logger_df: pd.DataFrame
            The logger file in pandas dataframe format.
        field_of_interest: str
            The field of interest that contains the clusters data.
        cluster_cell_type: list
            A list of cell type that are member of the cluster
        time_step: int (optional)
            The time step between time points. If not specified,
            the time step of the Morpheus file will be used.
        time_interval: list [start_time_point, end_time_point] (optional)
            You can use the time_interval to define the starting and ending
            point that is in the range of the original time interval of the
            simulation.
            If not specified, the original time interval will be used.
        time_symbol: str (default=t)
            The time symbol, as specified in the morpheus model.
        """
        super().__init__(name)
        self.field_of_interest = field_of_interest
        self.cluster_cell_type = cluster_cell_type
        self.time_step = time_step
        self.time_interval = time_interval
        self.time_symbol = time_symbol

    def __call__(self, loc):
        """
        Returns
        -------
        active_cell_count: int
            The total count of active infected cell, that means all cell that
            does not contribute to the cell-to-cell infection.
        """
        df = tsv_to_df(loc, "logger.csv")
        return get_count_cc_contributors_alltp(
            df, self.field_of_interest, self.cluster_cell_type,
            self.time_step, self.time_interval, self.time_symbol)


def get_count_cc_contributors_alltp(
        logger_df, field_of_interest, cluster_cell_type,
        time_step=None, time_interval=None, time_symbol='t'):
    """
    Get the total number of active infected cells in the grid at a
    specific time point
    """
    # check if time_step is specified, if not then calculate its value from the
    # distance between two time points from the morpheus logger file
    if time_step is None:
        time_steps = logger_df[time_symbol].unique()
        time_step = time_steps[1] - time_steps[0]
    inactive_cell_list = []
    # get the starting and ending time point
    if time_interval is None:
        start = logger_df[time_symbol].unique().tolist()[0]
        end = logger_df[time_symbol].unique().tolist()[-1]
    else:
        start = time_interval[0]
        end = time_interval[1]
    count_per_time_point = {}
    # filter the data to only consider the first time point
    filtered_data = logger_df[logger_df[time_symbol] == start]
    number_of_cells = filtered_data.shape[0]
    # calculate the edge size of the hexagonal grid
    edge = math.ceil(math.sqrt((number_of_cells - 1) / 3))
    # iterate through all cells in the specified time point
    for j in range(start, end + 1, time_step):
        filtered_data = logger_df[logger_df[time_symbol] == j]
        morpheus_logger_list = filtered_data[field_of_interest].tolist()
        full_list_of_cluster = []
        # iterate through all cells in the specified time point
        for index in range(len(morpheus_logger_list)):
            if (morpheus_logger_list[index] in cluster_cell_type) and \
                    (index + 1 not in full_list_of_cluster):
                index_list_of_cluster = []
                # find the neighbor cells list for a specific cell
                neighbor_cells_list = find_neighbor_list(index, edge)
                morpheus_neighbor_list = []
                # iterate through all neighbor cells
                for i in neighbor_cells_list:
                    morpheus_neighbor_list.append(morpheus_logger_list[i])
                index_list_of_cluster.append(index)
                # check if neighbor cells are CC active or not
                inactive_cell_list = check_neighbor_inactive_infected_cell(
                    morpheus_neighbor_list, index, inactive_cell_list,
                    cluster_cell_type)
                # find CC active cells
        active_cell_count = sum(
            i in cluster_cell_type for i in morpheus_logger_list)\
            - len(list(set(inactive_cell_list)))
        count_per_time_point[j] = active_cell_count
    return {"cc_contributors": list(count_per_time_point.values())}



"""
FitMultiCell
============

Fitting of multi-cellular models combining the tools morpheus and pyABC.
"""

from .version import __version__  # noqa: F401
from .model import *  # noqa: F403, F401
from .sumstat import *  # noqa: F403, F401
from .distance import *  # noqa: F403, F401

import os
import logging

# Set log level
try:
    loglevel = os.environ['FitMultiCell_LOG_LEVEL'].upper()
except KeyError:
    loglevel = 'INFO'
logger = logging.getLogger("FitMultiCell")
logger.setLevel(loglevel)
sh = logging.StreamHandler()
sh.setFormatter(logging.Formatter(
    '%(name)s %(levelname)s: %(message)s'))
logger.addHandler(sh)



import copy
import os
import pandas as pd
from typing import Union


def tsv_to_df(loc: str, file_: str = "logger.csv"):
    """
    Read in the morpheus logging file `file_`, which should be in
    tab-separated tsv format and defaults to "logger.csv", inside
    directory `loc`.

    Parameters
    ----------

    loc: str
        The simulations results folder.
    file_: str, optional (default="logger.csv")
        The logger file containing the results in tsv format.

    Returns
    -------

    data_frame: pandas.DataFrame
        A dataframe of the file contents.
    """
    data_file = os.path.join(loc, file_)
    data_frame = pd.read_csv(data_file, sep="\t")
    return data_frame


def split_list(alist, wanted_parts):
    """
    Take a set of lists based on the number of particles and turn that into a
    set of list based on the number of summary statistics.

    Parameters
    ----------
    alist: str
        list of particles readout
    wanted_parts: int
        should be equal to the number of summary statistics functions.
    Returns
    -------
        A list of lists, were the number of lists equal to the number of
        summary statistic functions.
    """
    return [alist[i::wanted_parts]
            for i in range(wanted_parts)]


def scaling_parameter(par_vals, par_scale: Union[str, dict]):
    """
    Rescaling parameter value to match the parameter space define in the
    prior distribution
    Parameters
    ----------
    par_vals: dict
        parameters name as a key and parameter value as a value of the dict.
    par_scale: str, or dict
        the scale to be used. It can be passed as a single value to be use for
         all parameters, or a dict to be applied to each parameter
         individually ,e.g., "log10".
    Returns
    -------
        a dictionary of the parameters value rescaled according to the
        par_scale.
    """
#    if isinstance(par_scale, str):
#        if par_scale == "log10":
#            par_vals.update((x, 10**y) for x, y in par_vals.items())
#        elif par_scale == "log2":
#            par_vals.update((x, 2**y) for x, y in par_vals.items())
#        elif par_scale == "linear":
#            par_vals
#        else:
#            raise ValueError(
#                f"The entered parameter scale 'par_scale'= {par_scale} is "
#                f"not supported")

#    if isinstance(par_scale, dict):
#        for key, scale in par_scale.items():
#            if scale == "log10":
#                par_vals[key] = 10 ** par_vals[key]
#            elif scale == "log2":
#                par_vals[key] = 2 ** par_vals[key]
#            elif scale == "linear":
#                continue
#            else:
#                raise ValueError(
#                    f"One of the entered parameter scale 'par_scale'=
#                       {scale} is "
#                    f"not supported")

    scaled_dict = copy.deepcopy(par_vals)
    if isinstance(par_scale, str):
        if par_scale == "log10":
            scaled_dict.update((x, 10**y) for x, y in par_vals.items())
        elif par_scale == "log2":
            scaled_dict.update((x, 2**y) for x, y in par_vals.items())
        elif par_scale == "lin":
            scaled_dict
        else:
            raise ValueError(
                f"The entered parameter scale 'par_scale'= {par_scale} is "
                f"not supported")
        return scaled_dict
    else:
        for key, val in par_scale.items():
            if val == "log10":
                scaled_dict[key] = 10 ** par_vals[key]
            elif val == "log2":
                scaled_dict[key] = 2 ** par_vals[key]
            elif val == "lin":
                scaled_dict[key] = par_vals[key]
            else:
                raise ValueError(
                    f"The entered parameter scale 'par_scale'= {par_scale} is "
                    f"not supported")
        return scaled_dict


def from_Morpheus_log_to_sumstat(loc: str):
    """
    Use the simulation output directly as a summary statistics.

    Parameters
    ----------
    loc: str
        The simulation folder.

    Returns
    -------
    sumstat: dict
        output file in a proper format
    """
    df = tsv_to_df(loc, "logger.csv")
    return df



import abc
from typing import Callable
import numpy as np
from ..util import tsv_to_df


class SumstatFun:
    """
    Base class for summary statistic functions working on
    morpheus models.

    Properties
    ----------

    name: str
        An identifier for his summary statistic. Used by pyabc
        to generate keys. The names of all summary statistics
        used jointly should be unique.
    """

    def __init__(self,
                 name: str,
                 class_name: str = None,
                 parameter_list: list = None):
        self.name = name
        self.class_name = class_name
        self.parameter_list = parameter_list

    @abc.abstractmethod
    def __call__(self, loc: str, output_file):
        """
        Parameters
        ----------

        loc: str
            The morpheus simulation folder.
        """
        if self.class_name and self.parameter_list is not None:

            self.parameter_list[0] = tsv_to_df(loc, output_file)
            func = getattr(self.class_name, self.name)
            result = func(*self.parameter_list)
            return result
        else:
            raise NotImplementedError()

    def create_own_sumstat_extractor(self):
        """
        Extract from the whole `sumstat` dictionary those entries
        that were generated by the present sumstat function.

        This is necessary since before the calculation of distances
        all sumstats are aggregated in a format which can also be
        stored.

        Returns
        -------

        own_sumstat_extractor: Callable[dict, dict]
            A function that extract the sumstats generated by the given
            sumstat fun based on the name attribute.
        """
        # extract all dependencies to avoid pickling issues
        name = self.name

        def own_sumstat_extractor(sumstat: dict):
            """
            Parameters
            ----------
            sumstat: dict
                The aggregated sumstat dictionary.

            Returns
            -------
            own_sumstat: dict
                The extracted values.
            """
            own_sumstat = {
                key: sumstat[key] for key in sumstat if
                key == name or key.startswith(name + "__")
            }
            return own_sumstat

        return own_sumstat_extractor

    def preprocess_distance_function(
            self, distance: Callable[[dict, dict], float]):
        """
        Preprocess the summary statistics to extract those values that
        were generated by the present summary statistics function.

        Parameters
        ----------
        distance: Callable[[dict, dict], float]
            The distance to apply to the given summary statistics.
            Note that subclasses may like to have a default distance to
            use for their data, such that the user does not need to
            specify anything at all.

        Returns
        -------
        preprocessed_distance: Callable[[dict, dict], float]
            A callable applying the `distance` to the own summary statistics
            only.
        """
        own_sumstat_extractor = self.create_own_sumstat_extractor()

        def preprocessed_distance(x, y):
            x = own_sumstat_extractor(x)
            y = own_sumstat_extractor(y)
            return distance(x, y)

        return preprocessed_distance


class IdSumstatFun(SumstatFun):

    def __init__(self, name: str = "IdSumstat"):
        if name != 'IdSumstat':
            name = name + '__' + 'IdSumstat'
        super().__init__(name)

    def __call__(self, loc: str, ignore: list,
                 output_file: str):
        """
        Parameters
        ----------
        loc: str
            location of the simulation output.
        ignore: list
            a list of keys to ignore for the simulation output. This is helpful
            to ignore information that cannot be avoid written by morpheus.
        """

        df = tsv_to_df(loc, output_file)
        dct = df.to_dict(orient='list')
        new_dct = dict()
        for key, val in dct.items():
            if key in ignore:
                continue
            _key = self.name + '__' + str(key)
            new_dct[_key] = np.array(val)
        return new_dct



from setuptools import setup, find_packages
import os


# extract version
with open(os.path.join(os.path.dirname(__file__),
                       "fitmulticell", "version.py")) as f:
    version = f.read().split("\n")[0].split("=")[-1].strip(' ').strip('"')


# read a file
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="fitmulticell",
    version=version,
    python_requires='>=3.7.1',
    install_requires=[
        'numpy>=1.16.2', 'scipy>=1.2.1', 'pandas>=0.25.0', 'matplotlib',
        'pyabc', 'imageio', 'PEtab_MS>=0.1.19'
    ],
    extra_requires={},
    packages=find_packages(exclude=["doc", "test*"]),
    author="The FitMultiCell developers",
    author_email='yannik.schaelte@gmail.com',
    platforms='all',
    url="https://gitlab.com/fitmulticell/fit",
    description="Tool for fitting multi-cellular models",
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
    ],
    license='BSD-3-Clause',
    keywords='fitmulticell, morpheus, pyabc, likelihood-free, '
             'approximate bayesian computation, abc',
)



from typing import List

from .base import SumstatFun
from ..util import tsv_to_df


class CellCountSumstatFun(SumstatFun):
    """
    Count the number of different cell types based on a field of interest.
    """

    def __init__(self,
                 field_of_interest: str,
                 cell_types_list: List,
                 time_step: int = None,
                 time_symbol: str = 't',
                 name: str = "CellCount"):
        """
        Parameters
        ----------
        field_of_interest: str
            The field of interest that we want to count its different cell
            types.
        cell_types_list: list
            A list of different type of cell that we want to count.
            E.g. [1, 2, 3].
        time_step: int (optional)
            The time step between time points. If not specified,
            the time step of the Morpheus file will be used.
        time_symbol: str (default=t)
            The time symbol, as specified in the morpheus model.
        """
        super().__init__(name)
        self.field_of_interest = field_of_interest
        self.cell_types_list = cell_types_list
        self.time_step = time_step
        self.time_symbol = time_symbol

    def __call__(self, loc: str):
        """
        Parameters
        ----------
        loc: str
            location of the simulation output.
        Returns
        -------
        A dictionary with the time points as keys, and for each as entry
        a pandas.DataFrame containing two columns: cell types and numbers of
        cells that have that type.
        """
        df = tsv_to_df(loc, "logger.csv")
        return count_cell_types(
            df, self.field_of_interest, self.cell_types_list,
            self.time_step, self.time_symbol)


def count_cell_types(logger_df,
                     field_of_interest,
                     cell_types_list,
                     time_step=None,
                     time_symbol='t'):
    """
    Count the number of different cell types based on a field of interest.
    """
    # set the time step value
    if not time_step:
        time_steps = logger_df[time_symbol].unique()
        time_step = time_steps[1] - time_steps[0]
    # define an empty dictionary for different time steps
    sumstat_dict = []
    # loop through the different time steps
    for i in range(len(cell_types_list)):
        sumstat_dict.append([])
    for j in range(
            logger_df[time_symbol].unique().tolist()[0],
            logger_df[time_symbol].unique().tolist()[-1] + 1,
            time_step):
        # takes the next time step
        filtered_data = logger_df[logger_df[time_symbol] == j]
        # time_series_counter = []
        # iterate through all different cell types

        for i, cell_type in enumerate(cell_types_list):
            # count the number of cells with a specific cell type
            sumstat_dict[i].append(
                filtered_data.groupby(
                    field_of_interest).size().get(cell_type, 0))
        # dict_time_series[j] = pd.DataFrame(
        #     {'cell_type': cell_types_list,
        #      'n_cells': time_series_counter}).set_index('cell_type')
    return {"cell_type": sumstat_dict}


def classify_based_on_value(
        logger_df, field_of_interest, value_of_interest,
        time_step=None, time_symbol='t'):
    """
    Classify the data of a field of interest based on a value of interest.

    Parameters
    ----------
    logger_df: pd.DataFrame
        The logger file in pandas dataframe format.
    field_of_interest: str
        The field of interest that we want to count its different cell types.
    value_of_interest: int
        the value of interest of which the classification will be based on.
    time_step: int (optional)
        The time step between time points. If not specified,
        the time step of the Morpheus file will be used.
    time_symbol: str (default=t)
        The time symbol, as specified in the morpheus model.
    Returns
    -------
    dict_time_series: dict
        A dictionary with cell type as a key, and the occurrences of
        the cell types as a value for different time intervals.
    """
    # set the time step value
    if not time_step:
        time_steps = logger_df[time_symbol].unique()
        time_step = time_steps[1] - time_steps[0]
    # get the length of the time step
    if field_of_interest not in logger_df:
        error_message = "Column name does not exist!"
        return error_message
    df_size = len(logger_df[time_symbol].unique().tolist())
    # define an empty dictionary for different time steps
    dict_time_series = {}
    series_obj = 0

    # loop through the different time steps
    for j in range(
            logger_df[time_symbol].unique().tolist()[0], df_size * time_step,
            time_step):
        # take the next time step
        filtered_data = logger_df[logger_df[time_symbol] == j]
        # print the RNA concentration
        for i in range(2):
            series_obj = filtered_data.apply(
                lambda x: True if x[field_of_interest] >= value_of_interest
                else False, axis=1)

            # count number of True in series
            num_of_rows = len(series_obj[series_obj is True].index)
        dict_time_series[j] = [len(series_obj) - num_of_rows, num_of_rows]
    return dict_time_series
